﻿using System;
using System.Runtime.InteropServices;
using System.Speech.Recognition;
using System.Speech.Recognition.SrgsGrammar;
using Microsoft.Office.Interop.Outlook;

namespace Palmmedia.Speech2Recipient
{
    /// <summary>
    /// Starts speech recognition when a new mail window is opened.
    /// All contacts from the adress book whose names are spoken, are added to the recipients list.
    /// </summary>
    internal class SpeechRecognizer : IDisposable
    {
        /// <summary>
        /// The current mail.
        /// </summary>
        private MailItem mail;        

        /// <summary>
        /// The current inspector.
        /// </summary>
        private Inspector inspector;

        /// <summary>
        /// The SpeechRecognitionEngine.
        /// </summary>
        private readonly SpeechRecognitionEngine speachRecognizerEngine = new SpeechRecognitionEngine();

        /// <summary>
        /// Reference to the inspectors.
        /// </summary>
        private Inspectors inspectors;

        /// <summary>
        /// Initializes a new instance of the <see cref="SpeechRegonizer"/> class.
        /// </summary>
        /// <param name="application">The application.</param>
        public SpeechRecognizer(Application application)
        {
            try
            {
                // Store a reference to the inspectors, otherwise events will stop firing
                this.inspectors = application.Inspectors;
                application.Inspectors.NewInspector += new InspectorsEvents_NewInspectorEventHandler(this.NewMailWindowOpening);

                this.SetupSpeechRecognitionEngine(application.ActiveExplorer().Session.GetDefaultFolder(OlDefaultFolders.olFolderContacts));         
            }
            finally
            {
                Marshal.ReleaseComObject(application);
                application = null;
            }            
        }

        /// <summary>
        /// Sets up the speech recognition engine.
        /// </summary>
        private void SetupSpeechRecognitionEngine(MAPIFolder contacts)
        {
            try
            {
                // Add all contacts
                SrgsOneOf srgsOneOf = new SrgsOneOf();

                for (int i = 1; i <= contacts.Items.Count; i++)
                {
                    var contact = contacts.Items[i] as ContactItem;

                    if (contact != null)
                    {
                        srgsOneOf.Add(new SrgsItem(contact.FirstName + " " + contact.LastName));
                        srgsOneOf.Add(new SrgsItem(contact.LastNameAndFirstName.Replace(",", "")));
                    }
                }

                // The rule containing the list of contacts
                SrgsRule rule = new SrgsRule("contacts");
                rule.Add(srgsOneOf);

                // The SrgsDocument containing the rule
                SrgsDocument doc = new SrgsDocument();
                doc.Rules.Add(rule);
                doc.Root = rule;

                // The grammer containing the SrgsDocument
                Grammar grammar = new Grammar(doc);

                // Init SpeechRecognitionEngine
                this.speachRecognizerEngine.SetInputToDefaultAudioDevice();
                this.speachRecognizerEngine.UnloadAllGrammars();
                this.speachRecognizerEngine.LoadGrammar(grammar);

                this.speachRecognizerEngine.SpeechRecognized += new EventHandler<SpeechRecognizedEventArgs>(this.RegonizedNewRecipient);
            }
            catch (System.Exception ex)
            {
                System.Windows.Forms.MessageBox.Show(ex.Message + " " + ex.StackTrace);
            }
            finally
            {
                Marshal.ReleaseComObject(contacts);
                contacts = null;
            }
        }

        /// <summary>
        /// Executed when a recipient is regonized.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.Speech.Recognition.SpeechRecognizedEventArgs"/> instance containing the event data.</param>
        void RegonizedNewRecipient(object sender, SpeechRecognizedEventArgs e)
        {
            this.mail.Recipients.Add(e.Result.Text);
        }

        /// <summary>
        /// Executed when new mail window is opening.
        /// </summary>
        /// <param name="inspector">The inspector.</param>
        private void NewMailWindowOpening(Inspector inspector)
		{
            this.inspector = inspector;
            ((InspectorClass)this.inspector).InspectorEvents_10_Event_Activate += new InspectorEvents_10_ActivateEventHandler(this.NewMailWindowOpened);
		}

        /// <summary>
        /// Executed when the mail window has finished opening.
        /// Speech recognition gets ativated now.
        /// </summary>
        private void NewMailWindowOpened()
        {
            this.mail = this.inspector.CurrentItem as MailItem;

            // Do no more listen to the activate event
            ((InspectorClass)this.inspector).InspectorEvents_10_Event_Activate -= new InspectorEvents_10_ActivateEventHandler(this.NewMailWindowOpened);

            if (this.mail != null)
            {
                // Now listen to deactivate event
                this.inspector.Deactivate += new InspectorEvents_10_DeactivateEventHandler(this.NewMailWindowDeactivated);

                this.speachRecognizerEngine.RecognizeAsync(RecognizeMode.Multiple);
            }
            else
            {
                Marshal.ReleaseComObject(this.inspector);
                this.inspector = null;
            }
        }

        /// <summary>
        /// Executed when mail window is deactivated.
        /// </summary>
        private void NewMailWindowDeactivated()
        {
            // Do no more listen to the deactivate event
            inspector.Deactivate -= new InspectorEvents_10_DeactivateEventHandler(this.NewMailWindowDeactivated);
            Marshal.ReleaseComObject(this.inspector);
            this.inspector = null;

            // Deactivate speech recognition
            if (speachRecognizerEngine != null)
            {
                this.speachRecognizerEngine.RecognizeAsyncCancel();
            }

            Marshal.ReleaseComObject(this.mail);
            this.mail = null;
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            if (this.speachRecognizerEngine != null)
            {
                this.speachRecognizerEngine.RecognizeAsyncCancel();
            }

            if (this.mail != null)
            {
                Marshal.ReleaseComObject(this.mail);
                this.mail = null;
            }

            if (this.inspector != null)
            {
                Marshal.ReleaseComObject(this.inspector);
                this.inspector = null;
            }

            if (this.inspectors != null)
            {
                Marshal.ReleaseComObject(this.inspectors);
                this.inspectors = null;
            }
        }
    }
}